"""
Utility methods for converting geographic locations to North/South B trace coordinates.
"""
import aacgmv2
from aurorax import Location
import datetime


def __calculate_btrace(geo_location: Location, dt) -> Location:
    # convert to magnetic coordinates
    mag_location = aacgmv2.convert_latlon(geo_location.lat,
                                          geo_location.lon,
                                          0.0,
                                          dt,
                                          method_code="G2A")

    # change magnetic latitude to other hemisphere
    mag_location = (
        mag_location[0] * -1.0,
        mag_location[1],
        mag_location[2]
    )

    # convert magnetic coordinates back to geographic
    btrace_aacgm = aacgmv2.convert_latlon(mag_location[0],
                                          mag_location[1],
                                          mag_location[2],
                                          dt,
                                          method_code="A2G")

    # return as Location object
    return Location(lat=btrace_aacgm[0], lon=btrace_aacgm[1])


def ground_geo_to_nbtrace(geo_location: Location, timestamp: datetime.datetime) -> Location:
    """
    Convert geographic location to North B-Trace geographic location.

    Attributes:
        geo_location: aurorax.Location object representing the geographic location.
        dt: datetime.datetime object representing the timestamp.

    Returns:
        North B-trace location as an aurorax.Location object.

    """
    # check if location is in northern hemisphere
    if (geo_location.lat >= 0):
        # northern hemisphere, north b-trace is the same as geographic location
        return geo_location

    # calculate South B-trace and return
    sbtrace = __calculate_btrace(geo_location, timestamp)
    return sbtrace


def ground_geo_to_sbtrace(geo_location: Location, timestamp: datetime.datetime) -> Location:
    """
    Convert geographic location to South B-Trace geographic location.

    Attributes:
        geo_location: aurorax.Location object representing the geographic location.
        dt: datetime.datetime object representing the timestamp.

    Returns:
        South B-trace location as an aurorax.Location object.

    """
    # check if location is in southern hemisphere
    if (geo_location.lat < 0):
        # southern hemisphere, south b-trace is the same as geographic location
        return geo_location

    # calculate North B-trace and return
    nbtrace = __calculate_btrace(geo_location, timestamp)
    return nbtrace



"""
AuroraX provides a conjunction module for finding conjunction events between ground and space instruments,
and between space instruments.
"""
import aurorax
import datetime
import humanize
import pprint
from pydantic import BaseModel
from typing import Dict, List

DEFAULT_CONJUNCTION_DISTANCE = 300


class Conjunction(BaseModel):
    """
    Conjunction data type.

    Attributes:
        conjunction_type: conjunction type "nbtrace" or "sbtrace".
        start: start datetime.datetime of conjunction event(s).
        end: end datetime.datetime of conjunction event(s).
        data_sources: aurorax.sources.DataSource sources in the conjunction.
        min_distance: minimum kilometre distance of conjunction event(s), float.
        max_distance: maximum kilometre distance of conjunction event(s), float.
        events: list of dictionaries containing details of individual conjunction events.

    """
    conjunction_type: str
    start: datetime.datetime
    end: datetime.datetime
    data_sources: List[aurorax.sources.DataSource]
    min_distance: float
    max_distance: float
    events: List[Dict]

    def __str__(self) -> str:
        """
        String method.

        Returns:
            String format of Conjunction object.

        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation.

        Returns:
            Object representation of Conjunction object.

        """
        return pprint.pformat(self.__dict__)


class Search():
    """
    Class representing an AuroraX conjunctions search.

    Attributes:
        start: start datetime.datetime timestamp of the search.
        end: end datetime.datetime timestamp of the search.
        ground: List of ground instrument search parameters. See examples for usage.
            e.g. [
                {
                    "programs": ["themis-asi"],
                    "platforms": ["gillam", "rabbit lake"],
                    "instrument_types": ["RGB"]
                }
            ]
        space: List of one or more space instrument search parameters. See examples for usage.
            e.g. [
                {
                    "programs": ["themis-asi", "swarm"],
                    "platforms": ["themisa", "swarma"],
                    "instrument_types": ["footprint"]
                }
            ]
        conjunction_types: list of conjunction types, defaults to ["nbtrace"].
        max_distances: dictionary of Dict[str, float] ground-space and space-space maximum distances for conjunctions.
            default_distance will be used for any ground-space and space-space maximum distances not specified.
            See examples for usage.
            e.g. distances = {
                "ground1-ground2": None,
                "ground1-space1": 500,
                "ground1-space2": 500,
                "ground2-space1": 500,
                "ground2-space2": 500,
                "space1-space2": None
            }
        default_distance: default maximum distance in kilometers for conjunction.
            Used when max distance is not specified for any ground-space and space-space instrument pairs.
        epoch_search_precision: the time precision to which conjunctions are calculated. Can be 30 or 60
            seconds. Defaults to 60 seconds.
        request: aurorax.AuroraXResponse object returned when the search is executed.
        request_id: unique AuroraX string ID assigned to the request.
        request_url: unique AuroraX URL string assigned to the request.
        executed: boolean, gets set to True when the search is executed.
        completed: boolean, gets set to True when the search is checked to be finished.
        data_url: URL string where data is accessed.
        query: dictionary of values sent for the search query.
        status: dictionary of status updates.
        data: list of aurorax.conjunctions.Conjunction objects returned.
        logs: list of logging messages from the API.

        Returns:
            An aurorax.conjunctions.Search object.

    """

    def __init__(self, start: datetime.datetime, end: datetime.datetime,
                 ground: List[Dict], space: List[Dict], conjunction_types: List[str] = ["nbtrace"],
                 max_distances: Dict[str, float] = None, default_distance: float = DEFAULT_CONJUNCTION_DISTANCE,
                 epoch_search_precision: int = 60):

        self.request = None
        self.request_id = ""
        self.request_url = ""
        self.executed = False
        self.completed = False
        self.data_url = ""
        self.query = {}
        self.status = {}
        self.data: List[Conjunction] = []
        self.logs = []

        self.start = start
        self.end = end
        self.ground = ground
        self.space = space
        self.conjunction_types = conjunction_types
        self.max_distances = max_distances if max_distances else {}
        self.default_distance = default_distance
        self.epoch_search_precision = epoch_search_precision

    def __str__(self):
        """
        String method.

        Returns:
            String format of Conjunction Search object.

        """
        return self.__repr__()

    def __repr__(self):
        """
        Object representation.

        Returns:
            Object representation of Conjunction Search object.

        """
        return pprint.pformat(self.__dict__)

    def _set_max_distances(self):
        ground_sources_len = len(self.ground)
        space_sources_len = len(self.space)

        # check for ground-space and space-ground distances
        for g in range(1, ground_sources_len + 1):
            for s in range(1, space_sources_len + 1):
                if (f"ground{g}-space{s}" not in self.max_distances
                        and f"space{s}-ground{g}" not in self.max_distances):
                    self.max_distances[f"ground{g}-space{s}"] = self.default_distance

        # check for space-space distances
        for s in range(1, space_sources_len + 1):
            for s2 in range(s + 1, space_sources_len + 1):
                if (s != s2 and f"space{s}-space{s2}" not in self.max_distances
                        and f"space{s2}-space{s}" not in self.max_distances):
                    self.max_distances[f"space{s}-space{s2}"] = self.default_distance

    def _check_num_criteria_blocks(self):
        ground_criteria_len = len(self.ground)
        space_criteria_len = len(self.space)

        if ground_criteria_len + space_criteria_len > 10:
            raise aurorax.exceptions.AuroraXBadParametersException(
                "Number of ground + space criteria blocks exceeds 10. Please use a smaller search criteria."
            )

        return

    def execute(self):
        """
        Initiate conjunction search request.

        Raises:
            aurorax.exceptions.AuroraXBadParametersException: too many criteria blocks.
        """
        # check number of criteria blocks
        self._check_num_criteria_blocks()

        # set up request
        url = aurorax.api.urls.conjunction_search_url
        self._set_max_distances()
        post_data = {
            "start": self.start.strftime("%Y-%m-%dT%H:%M:%S"),
            "end": self.end.strftime("%Y-%m-%dT%H:%M:%S"),
            "ground": self.ground,
            "space": self.space,
            "conjunction_types": self.conjunction_types,
            "max_distances": self.max_distances,
            "epoch_search_precision": self.epoch_search_precision if self.epoch_search_precision in [30, 60] else 60,
        }
        self.query = post_data

        # do request
        req = aurorax.AuroraXRequest(
            method="post", url=url, body=post_data, null_response=True)
        res = req.execute()

        # set request ID, request_url, executed
        self.executed = True
        if res.status_code == 202:
            # request successfully dispatched
            self.executed = True
            self.request_url = res.request.headers["location"]
            self.request_id = self.request_url.rsplit("/", 1)[-1]

        self.request = res

    def update_status(self, status: Dict = None) -> None:
        """
        Update the status of this conjunctions search.

        Attributes:
            status: retrieved status dictionary (include to avoid requesting it from the API again), defaults to None.
        """
        # get the status if it isn't passed in
        if status is None:
            status = aurorax.requests.get_status(self.request_url)

        # update request status by checking if data URI is set
        if status["search_result"]["data_uri"] is not None:
            self.completed = True
            self.data_url = f'{aurorax.api.urls.base_url}{status["search_result"]["data_uri"]}'

        # set class variable "status" and "logs"
        self.status = status
        self.logs = status["logs"]

    def check_for_data(self) -> bool:
        """
        Check to see if data is available for this conjunctions search request.

        Returns:
            True if data is available, else False.
        """
        self.update_status()

        return self.completed

    def get_data(self) -> None:
        """
        Retrieve the data available for this conjunctions search request.
        """
        if not self.completed:
            print("No data available, update status or check for data first")
            return

        url = self.data_url
        raw_data = aurorax.requests.get_data(url)
        self.data = [Conjunction(**c) for c in raw_data]

    def wait(self, poll_interval: float = aurorax.requests.STANDARD_POLLING_SLEEP_TIME, verbose: bool = False) -> None:
        """
        Block and wait until the request is complete and data is available for retrieval.

        Attributes:
            poll_interval: time in seconds to wait between polling attempts, defaults
                to aurorax.requests.STANDARD_POLLING_SLEEP_TIME.
            verbose: output poll times, defaults to False.

        """
        url = aurorax.api.urls.conjunction_request_url.format(self.request_id)
        self.update_status(aurorax.requests.wait_for_data(
            url, poll_interval=poll_interval, verbose=verbose))

    def cancel(self, wait: bool = False, verbose: bool = False,
               poll_interval: float = aurorax.requests.STANDARD_POLLING_SLEEP_TIME) -> int:
        """
        Cancel the conjunction search request at the API. This method returns asynchronously by default.

        Attributes:
            wait: set to True to block until the cancellation request has been completed. This may take several minutes.
            verbose: when wait=True, output poll times, defaults to False.
            poll_interval: when wait=True, seconds to wait between polling calls, defaults to STANDARD_POLLING_SLEEP_TIME.

        Returns:
            1 on success.

        Raises:
            aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.
            aurorax.exceptions.AuroraXUnauthorizedException: invalid API key for this operation.

        """
        url = aurorax.api.urls.conjunction_request_url.format(self.request_id)
        return aurorax.requests.cancel(url, wait=wait, poll_interval=poll_interval, verbose=verbose)


def search_async(start: datetime.datetime,
                 end: datetime.datetime,
                 ground: List[Dict],
                 space: List[Dict],
                 conjunction_types: List[str] = ["nbtrace"],
                 max_distances: Dict[str, float] = {},
                 default_distance: float = DEFAULT_CONJUNCTION_DISTANCE,
                 epoch_search_precision: int = 60) -> Search:
    """
    Submit a request for a conjunctions search, return asynchronously.

    Attributes:
        start: start datetime.datetime timestamp of the search.
        end: end datetime.datetime timestamp of the search.
        ground: List of ground instrument search parameters. See examples for usage.
            e.g. [
                {
                    "programs": ["themis-asi"],
                    "platforms": ["gillam", "rabbit lake"],
                    "instrument_types": ["RGB"]
                }
            ]
        space: List of one or more space instrument search parameters. See examples for usage.
            e.g. [
                {
                    "programs": ["themis-asi", "swarm"],
                    "platforms": ["themisa", "swarma"],
                    "instrument_types": ["footprint"]
                }
            ]
        conjunction_types: list of conjunction types, defaults to ["nbtrace"].
        max_distances: dictionary of Dict[str, float] ground-space and space-space maximum distances for conjunctions.
            default_distance will be used for any ground-space and space-space maximum distances not specified.
            See examples for usage.
            e.g. distances = {
                "ground1-ground2": None,
                "ground1-space1": 500,
                "ground1-space2": 500,
                "ground2-space1": 500,
                "ground2-space2": 500,
                "space1-space2": None
            }
        default_distance: default maximum distance in kilometers for conjunction.
            Used when max distance is not specified for any ground-space and space-space instrument pairs.
        epoch_search_precision: the time precision to which conjunctions are calculated. Can be 30 or 60
            seconds. Defaults to 60 seconds.

    Returns:
        An aurorax.conjunctions.Search object.
    """
    s = Search(start=start, end=end, ground=ground, space=space, conjunction_types=conjunction_types,
               max_distances=max_distances, default_distance=default_distance, epoch_search_precision=epoch_search_precision)
    s.execute()

    return s


def search(start: datetime.datetime,
           end: datetime.datetime,
           ground: List[Dict],
           space: List[Dict],
           conjunction_types: List[str] = ["nbtrace"],
           max_distances: Dict[str, float] = {},
           default_distance: float = DEFAULT_CONJUNCTION_DISTANCE,
           verbose: bool = False,
           poll_interval: float = aurorax.requests.STANDARD_POLLING_SLEEP_TIME,
           epoch_search_precision: int = 60) -> Search:
    """
    Search for conjunctions and block until results are returned.

    Attributes:
        start: start datetime.datetime timestamp of the search.
        end: end datetime.datetime timestamp of the search.
        ground: List of ground instrument search parameters. See examples for usage.
            e.g. [
                {
                    "programs": ["themis-asi"],
                    "platforms": ["gillam", "rabbit lake"],
                    "instrument_types": ["RGB"]
                }
            ]
        space: List of one or more space instrument search parameters. See examples for usage.
            e.g. [
                {
                    "programs": ["themis-asi", "swarm"],
                    "platforms": ["themisa", "swarma"],
                    "instrument_types": ["footprint"]
                }
            ]
        conjunction_types: list of conjunction types, defaults to ["nbtrace"].
        max_distances: dictionary of Dict[str, float] ground-space and space-space maximum distances for conjunctions.
            default_distance will be used for any ground-space and space-space maximum distances not specified.
            See examples for usage.
            e.g. distances = {
                "ground1-ground2": None,
                "ground1-space1": 500,
                "ground1-space2": 500,
                "ground2-space1": 500,
                "ground2-space2": 500,
                "space1-space2": None
            }
        default_distance: default maximum distance in kilometers for conjunction.
            Used when max distance is not specified for any ground-space and space-space instrument pairs.
        verbose: boolean to show the progress of the request using the request log, defaults to False.
        poll_interval: seconds to wait between polling calls, defaults to aurorax.requests.STANDARD_POLLING_SLEEP_TIME.

    Returns:
        An aurorax.conjunctions.Search object.

    """
    # create a Search object
    s = Search(start, end, ground, space, conjunction_types,
               max_distances, default_distance, epoch_search_precision)
    if verbose:
        print(f"[{datetime.datetime.now()}] Search object created")

    # execute the search
    s.execute()
    if verbose:
        print(f"[{datetime.datetime.now()}] Request submitted")
        print(f"[{datetime.datetime.now()}] Request ID: {s.request_id}")
        print(
            f"[{datetime.datetime.now()}] Request details available at: {s.request_url}")
        print(f"[{datetime.datetime.now()}] Waiting for data ...")

    s.wait(poll_interval=poll_interval, verbose=verbose)

    # get the data
    if verbose:
        print(f"[{datetime.datetime.now()}] Retrieving data ...")

    s.get_data()

    if verbose:
        print(f'[{datetime.datetime.now()}] Retrieved \
        {humanize.filesize.naturalsize(s.status["search_result"]["file_size"])} of data containing \
        {s.status["search_result"]["result_count"]} records')

    return s



"""
The availability module provides the means to quickly determine the availability of
desired ephemeris or data product records in AuroraX.
"""
import aurorax
import datetime
import pprint
from pydantic import BaseModel
from typing import Dict, List


class AvailabilityResult(BaseModel):
    """
    Availability result data type.

    Attributes:
        data_source: aurorax.sources.DataSource object that the ephemeris record is associated with.
        available_data_products: data product availability dictionary of shape {"YYYY-MM-DD": <# of records>}.
        available_ephemeris: ephemeris availability dictionary of shape {"YYYY-MM-DD": <# of records>}.

    """
    data_source: aurorax.sources.DataSource
    available_data_products: Dict[str, int] = None
    available_ephemeris: Dict[str, int] = None

    def __str__(self) -> str:
        """
        String method.

        Returns:
            String format of AvailabilityResult.

        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation.

        Returns:
            Object representation of AvailabilityResult.

        """
        return pprint.pformat(self.__dict__)


def ephemeris(start: datetime.date,
              end: datetime.date,
              program: str = None,
              platform: str = None,
              instrument_type: str = None,
              source_type: str = None,
              owner: str = None,
              format: str = "basic_info",
              slow: bool = False) -> List[AvailabilityResult]:
    """
    Retrieve information about the number of existing ephemeris records.

    Attributes:
        start: start datetime.date.
        end: end datetime.date.
        program: program string name to filter sources by, defaults to None.
        platform: platform string name to filter sources by, defaults to None.
        instrument_type: instrument type string to filter sources by, defaults to None.
        source_type: source type string to filter sources by (heo, leo, lunar, or ground), defaults to None.
        owner: owner email address string to filter sources by, defaults to None.
        format: the format of the data sources returned (identifier_only, basic_info,
            full_record), defaults to "basic_info".
        slow: query the data directly (slower) for more accurate results, defaults to false.

    Returns:
        A list of aurorax.availability.AvailabilityResult objects.

    """
    # set parameters
    params = {
        "start": start.strftime("%Y-%m-%d"),
        "end": end.strftime("%Y-%m-%d"),
        "program": program,
        "platform": platform,
        "instrument_type": instrument_type,
        "source_type": source_type,
        "owner": owner,
        "format": format,
        "slow": slow,
    }

    # do request
    req = aurorax.AuroraXRequest(
        method="get", url=aurorax.api.urls.ephemeris_availability_url, params=params)
    res = req.execute()

    # return
    return [AvailabilityResult(**av) for av in res.data]


def data_products(start: datetime,
                  end: datetime,
                  program: str = None,
                  platform: str = None,
                  instrument_type: str = None,
                  source_type: str = None,
                  owner: str = None,
                  format: str = "basic_info",
                  slow: bool = False) -> List[AvailabilityResult]:
    """
    Retrieve information about the number of existing data product records.

    Attributes:
        start: start datetime.date.
        end: end datetime.date.
        program: program string name to filter sources by, defaults to None.
        platform: platform string name to filter sources by, defaults to None.
        instrument_type: instrument type string to filter sources by, defaults to None.
        source_type: source type string to filter sources by (heo, leo, lunar, or ground), defaults to None.
        owner: owner email address string to filter sources by, defaults to None.
        format: the format of the data sources returned (identifier_only, basic_info,
            full_record), defaults to "basic_info".
        slow: query the data directly (slower) for more accurate results, defaults to false.

    Returns:
        A list of aurorax.availability.AvailabilityResult objects.

    """
    # set parameters
    params = {
        "start": start.strftime("%Y-%m-%d"),
        "end": end.strftime("%Y-%m-%d"),
        "program": program,
        "platform": platform,
        "instrument_type": instrument_type,
        "source_type": source_type,
        "owner": owner,
        "format": format,
        "slow": slow,
    }

    # do request
    req = aurorax.AuroraXRequest(
        method="get", url=aurorax.api.urls.data_products_availability_url, params=params)
    res = req.execute()

    # return
    return [AvailabilityResult(**av) for av in res.data]



"""
AuroraX data sources are unique instruments that produce ephemeris or data product records.
"""
import aurorax
import datetime
import pprint
from pydantic import BaseModel
from typing import List, Dict


class DataSource(BaseModel):
    """
    Data source data type.

    Attributes:
        identifier: an integer unique to the data source.
        program: a string representing the data source program.
        platform: a string representing the data source platform.
        instrument_type: a string representing the data source instrument type.
        source_type: a string representing the data source type.
        display_name: a string representing the data source's proper display name.
        metadata: a dictionary of metadata properties.
        owner: a string representing the data source's owner in AuroraX.
        maintainers: a list of strings representing the email addresses of AuroraX accounts
            that can alter this data source and its associated records.
        ephemeris_metadata_schema: a list of dictionaries capturing the metadata keys and values
            that can appear in ephemeris records associated with the data source.
        data_product_metadata_schema: a list of dictionaries capturing the metadata keys and values
            that can appear in data product records associated with the data source.

    """
    identifier: int = None
    program: str = None
    platform: str = None
    instrument_type: str = None
    source_type: str = None
    display_name: str = None
    metadata: Dict = None
    owner: str = None
    maintainers: List[str] = None
    ephemeris_metadata_schema: List[Dict] = None
    data_product_metadata_schema: List[Dict] = None

    def __str__(self) -> str:
        """
        String method.

        Returns:
            String format of DataSource object.

        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation.

        Returns:
            Object representation of DataSource object.

        """
        return pprint.pformat(self.__dict__)


class DataSourceStatistics(BaseModel):
    """
    Data type for data source statistics.

    Attributes:
        data_source: the data source the statistics are associated with
        earliest_ephemeris_loaded: datetime.datetime of the earliest ephemeris record
        latest_ephemeris_loaded: datetime.datetime of the latest ephemeris record
        ephemeris_count: total number of ephemeris records for this data source
        earliest_data_product_loaded: datetime.datetime of the earliest data_product record
        latest_data_product_loaded: datetime.datetime of the latest data product record
        data_product_count: total number of ephemeris records for this data source
    """
    data_source: DataSource
    earliest_ephemeris_loaded: datetime.datetime = None
    latest_ephemeris_loaded: datetime.datetime = None
    ephemeris_count: int
    earliest_data_product_loaded: datetime.datetime = None
    latest_data_product_loaded: datetime.datetime = None
    data_product_count: int

    def __str__(self) -> str:
        """
        String method.

        Returns:
            String format of DataSource object.

        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation.

        Returns:
            Object representation of DataSource object.

        """
        return pprint.pformat(self.__dict__)


def list(order: str = "identifier", format: str = "basic_info") -> List[DataSource]:
    """
    Retrieve all data source records.

    Attributes:
        order: string value to order results by (identifier, program, platform,
            instrument_type, display_name, owner), defaults to "identifier".
        format: string record format, defaults to "basic_info".

    Returns:
        A list of AuroraX DataSource objects.

    Raises:
        aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.
    """
    params = {
        format: format
    }
    # make request
    req = aurorax.AuroraXRequest(
        method="get", url=aurorax.api.urls.data_sources_url, params=params if format else None)
    res = req.execute()

    # order results
    res.data = sorted(res.data, key=lambda x: x[order])

    # return
    return [DataSource(**source) for source in res.data]


def get(program: str,
        platform: str,
        instrument_type: str,
        format: str = "basic_info") -> DataSource:
    """
    Retrieve a specific data source record.

    Attributes:
        program: the string name of the program.
        platform: the string name of the platform.
        instrument_type: the string name of the instrument type.
        format: record format, defaults to "basic_info".

    Returns:
        An AuroraX DataSource object matching the input parameters.

    Raises:
        aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.
        aurorax.exceptions.AuroraXNotFoundException: source not found.

    """
    # make request
    params = {
        "program": program,
        "platform": platform,
        "instrument_type": instrument_type,
        "format": format,
    }
    req = aurorax.AuroraXRequest(
        method="get", url=aurorax.api.urls.data_sources_url, params=params)
    res = req.execute()

    # set results to the first thing
    if (len(res.data) == 1):
        res.data = res.data[0]

        return DataSource(**res.data)
    else:
        raise aurorax.AuroraXNotFoundException("Data source not found")


def get_using_filters(program: str = None,
                      platform: str = None,
                      instrument_type: str = None,
                      source_type: str = None,
                      owner: str = None,
                      format: str = "basic_info",
                      order: str = "identifier",) -> List[DataSource]:
    """
    Retrieve all data source records matching a filter.

    Attributes:
        program: the string name of the program.
        platform: the string name of the platform.
        instrument_type: the string name of the instrument type.
        source_type: the string name of the data source type.
        owner: the AuroraX data source owner's email address.
        format: record format, defaults to "basic_info".
        order: string value to order results by (identifier, program, platform,
            instrument_type, display_name, owner), defaults to "identifier".

    Returns:
        A list of AuroraX DataSource objects matching the filter parameters.

    Raises:
        aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.

    """
    # make request
    params = {
        "program": program,
        "platform": platform,
        "instrument_type": instrument_type,
        "source_type": source_type,
        "owner": owner,
        "format": format,
    }
    req = aurorax.AuroraXRequest(
        method="get", url=aurorax.api.urls.data_sources_url, params=params)
    res = req.execute()

    # order results
    res.data = sorted(res.data, key=lambda x: x[order])

    # return
    if len(res.data):
        return [DataSource(**ds) for ds in res.data]
    else:
        return []


def get_using_identifier(identifier: int, format: str = "basic_info") -> DataSource:
    """
    Retrieve data source record matching an identifier.

    Attributes:
        identifier: an integer unique to the data source.
        format: record format, defaults to "basic_info".

    Returns:
        An AuroraX DataSource object matching the input identifier.

    Raises:
        aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.

    """
    # make request
    params = {
        "format": format,
    }
    url = "%s/%d" % (aurorax.api.urls.data_sources_url, identifier)
    req = aurorax.AuroraXRequest(method="get", url=url, params=params)
    res = req.execute()

    # return
    return DataSource(**res.data)


def get_stats(identifier: int,
              format: str = "basic_info",
              slow: bool = False) -> Dict:
    """
    Retrieve statistics for a data source.

    Attributes:
        identifier: an integer unique to the data source.
        format: record format, defaults to "basic_info".
        slow: a boolean indicating whether to use slow method which gets most up-to-date stats info.

    Returns:
        A dictionary of statistics information about the data source.

    Raises:
        aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
        aurorax.exceptions.AuroraXNotFoundException: data source not found.
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.

    """
    # make request
    params = {
        "format": format,
        "slow": slow,
    }
    url = "%s/%d/stats" % (aurorax.api.urls.data_sources_url, identifier)
    req = aurorax.AuroraXRequest(method="get", url=url, params=params)
    res = req.execute()

    # return
    return DataSourceStatistics(**res.data)


def add(data_source: DataSource) -> DataSource:
    """
    Add a new data source to AuroraX.

    Attributes:
        data_source: the fully defined AuroraX DataSource object to add.

    Returns:
        The newly created AuroraX DataSource object.

    Raises:
        aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.
        aurorax.exceptions.AuroraXDuplicateException: duplicate data source, already exists.

    """

    # do request
    request_data = {
        "program": data_source.program,
        "platform": data_source.platform,
        "instrument_type": data_source.instrument_type,
        "source_type": data_source.source_type,
        "display_name": data_source.display_name,
        "ephemeris_metadata_schema": data_source.ephemeris_metadata_schema,
        "data_product_metadata_schema": data_source.data_product_metadata_schema,
        "metadata": data_source.metadata
    }

    if (data_source.identifier is not None):
        request_data["identifier"] = data_source.identifier

    req = aurorax.AuroraXRequest(
        method="post", url=aurorax.api.urls.data_sources_url, body=request_data)
    res = req.execute()

    # evaluate response
    if (res.status_code == 409):
        raise aurorax.AuroraXDuplicateException(
            "%s - %s" % (res.data["error_code"], res.data["error_message"]))

    # return
    try:
        return DataSource(**res.data)
    except Exception:
        raise aurorax.AuroraXException("Could not create data source.")


def delete(identifier: int) -> int:
    """
    Delete a data source from AuroraX.

    Attributes:
        identifier: an integer unique to the data source.

    Returns:
        1 on success.

    Raises:
        aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.
        aurorax.exceptions.AuroraXNotFoundException: data source not found.
        aurorax.exceptions.AuroraXConflictException: conflict of some type.

    """
    # do request
    url = "%s/%d" % (aurorax.api.urls.data_sources_url, identifier)
    req = aurorax.AuroraXRequest(method="delete", url=url, null_response=True)
    res = req.execute()

    # evaluate response
    if (res.status_code == 400):
        raise aurorax.AuroraXBadParametersException(
            "%s - %s" % (res.data["error_code"], res.data["error_message"]))
    elif (res.status_code == 409):
        raise aurorax.AuroraXConflictException(
            "%s - %s" % (res.data["error_code"], res.data["error_message"]))

    # return
    return 1


def update(data_source: DataSource) -> DataSource:
    """
    Update a data source in AuroraX.

    This operation will fully replace the data source with the data_source argument passed in.
    Please make sure data_source is complete. Refer to examples for usage.

    Attributes:
        data_source: the fully defined AuroraX DataSource object to update.

    Returns:
        The updated AuroraX DataSource object.

    Raises:
        aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.
        aurorax.exceptions.AuroraXNotFoundException: data source not found.
        aurorax.exceptions.AuroraXBadParametersException: missing parameters.

    """
    if not all([data_source.identifier, data_source.program, data_source.platform,
                data_source.instrument_type, data_source.source_type, data_source.display_name]):
        raise aurorax.AuroraXBadParametersException(
            "One or more required data source fields are missing. Update operation aborted.")

    # set URL
    url = f"{aurorax.api.urls.data_sources_url}/{data_source.identifier}"

    # make request to update the data source passed in
    req = aurorax.AuroraXRequest(method="put", url=url, body=data_source)
    req.execute()

    # return
    try:
        return aurorax.sources.get(data_source.program, data_source.platform, data_source.instrument_type, "full_record")
    except Exception:
        raise aurorax.AuroraXException("Could not update data source.")


def partial_update(identifier: int,
                   program: str = None,
                   platform: str = None,
                   instrument_type: str = None,
                   source_type: str = None,
                   display_name: str = None,
                   metadata: Dict = None,
                   owner: str = None,
                   maintainers: List[str] = None,
                   ephemeris_metadata_schema: List[Dict] = None,
                   data_product_metadata_schema: List[Dict] = None) -> DataSource:
    """
    Partially update a data source in AuroraX.

    Omitted fields are ignored in the update. Refer to examples for usage.

    Attributes:
        identifier: an integer unique to the data source.
        program: a string representing the data source program.
        platform: a string representing the data source platform.
        instrument_type: a string representing the data source instrument type.
        source_type: a string representing the data source type.
        display_name: a string representing the data source's proper display name.
        metadata: a dictionary of metadata properties.
        owner: a string representing the data source's owner in AuroraX.
        maintainers: a list of strings representing the email addresses of AuroraX accounts
            that can alter this data source and its associated records.
        ephemeris_metadata_schema: a list of dictionaries capturing the metadata keys and values
            that can appear in ephemeris records associated with the data source.
        data_product_metadata_schema: a list of dictionaries capturing the metadata keys and values
            that can appear in data product records associated with the data source.

    Returns:
        The updated AuroraX DataSource object.

    Raises:
        aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.
        aurorax.exceptions.AuroraXNotFoundException: data source not found.
        aurorax.exceptions.AuroraXBadParametersException: missing parameters.

    """
    if not identifier:
        raise aurorax.AuroraXBadParametersException(
            "Required identifier field is missing. Update operation aborted.")

    # create a DataSource
    ds = DataSource(identifier=identifier,
                    program=program,
                    platform=platform,
                    instrument_type=instrument_type,
                    source_type=source_type,
                    display_name=display_name,
                    metadata=metadata,
                    owner=owner,
                    maintainers=maintainers,
                    ephemeris_metadata_schema=ephemeris_metadata_schema,
                    data_product_metadata_schema=data_product_metadata_schema)

    # set URL
    url = f"{aurorax.api.urls.data_sources_url}/{ds.identifier}"

    # make request to update the data source passed in
    req = aurorax.AuroraXRequest(method="patch", url=url, body=ds)
    req.execute()

    # return
    try:
        return aurorax.sources.get_using_identifier(ds.identifier, "full_record")
    except Exception:
        raise aurorax.AuroraXException("Could not update data source.")



"""
The API module contains classes and methods used throughout PyAuroraX for API interaction.
"""
import aurorax
import json
import pprint
from pydantic import BaseModel
import requests
from typing import Optional, Dict, Any, List, Union
from ._internal.util import json_converter

# endpoint URLs
DEFAULT_URL_BASE = "https://api.aurorax.space"

# reqest globals
DEFAULT_RETRIES = 2
REQUEST_HEADERS = {
    "accept": "application/json",
    "Content-Type": "application/json"
}
API_KEY_HEADER_NAME = "x-aurorax-api-key"

# private dynamic globals
__api_key = ""


def get_api_key() -> str:
    """
    Returns the currently set API key for the module.

    Returns:
        Current API key string.
    """
    return __api_key


def authenticate(api_key: str) -> None:
    """
    Set authentication values for use with subsequent queries.

    Attributes:
        api_key: AuroraX API key string.

    """

    # set the global variable
    global __api_key
    __api_key = api_key


class AuroraXResponse(BaseModel):
    request: Any
    data: Any
    status_code: int

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return pprint.pformat(self.__dict__)


class AuroraXRequest(BaseModel):
    url: str
    method: str
    params: Optional[Dict] = {}
    body: Union[Optional[Dict], Optional[List]] = {}
    headers: Optional[Dict] = {}
    null_response: Optional[bool] = False

    def __merge_headers(self):
        # set initial headers
        all_headers = {
            "accept": "application/json",
            "Content-Type": "application/json"
        }

        # add headers passed into the class
        for key, value in self.headers.items():
            all_headers[key] = value

        # add api key
        api_key = get_api_key()
        if api_key:
            all_headers[API_KEY_HEADER_NAME] = api_key

        # return
        return all_headers

    def execute(self, limited_evaluation: bool = False) -> AuroraXResponse:
        """
        Execute an AuroraX request.

        Attributes:
            limited_evaluation: set this to True if you don't want to evaluate the response outside of
                the retry mechanism, defaults to False.

        Returns:
            An AuroraXResponse object.

        Raises:
            aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
            aurorax.exceptions.AuroraXNotFoundException: requested resource was not found.
            aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected content error.
            aurorax.exceptions.AuroraXUnauthorizedException: invalid API key for this operation.

        """
        # sanitize data
        body_santized = json.dumps(self.body, default=json_converter)

        # make request
        req = requests.request(self.method,
                               self.url,
                               headers=self.__merge_headers(),
                               params=self.params,
                               data=body_santized)

        # retry request if needed
        for i in range(0, aurorax.api.DEFAULT_RETRIES):
            if (req.status_code == 500 and "text/plain" in req.headers["Content-Type"]):
                if (i == (aurorax.api.DEFAULT_RETRIES - 1)):
                    raise aurorax.AuroraXMaxRetriesException("%s (%s)" % (req.content.decode(),
                                                                          req.status_code))
                req = requests.request(self.method,
                                       self.url,
                                       headers=self.__merge_headers(),
                                       params=self.params,
                                       json=self.body,
                                       data=body_santized)
            else:
                break

        # check if authorization worked
        if (req.status_code == 401):
            raise aurorax.AuroraXUnauthorizedException("%s %s" % (req.status_code,
                                                                  req.json()["error_message"]))

        if (req.status_code == 404):
            raise aurorax.AuroraXNotFoundException("%s %s" % (req.status_code,
                                                              req.json()["error_message"]))

        # check if we only want to do limited evaluation
        if (limited_evaluation is True):
            res = AuroraXResponse(request=req, data=None,
                                  status_code=req.status_code)
            return res

        # check content type
        if (self.null_response is False):
            if (req.headers["Content-Type"] == "application/json"):
                response_data = req.json()
            else:
                raise aurorax.AuroraXUnexpectedContentTypeException("%s (%s)" % (req.content.decode(),
                                                                                 req.status_code))
        else:
            if (req.status_code != 200 and req.status_code != 201 and req.status_code != 202 and req.status_code != 204):
                response_data = req.json()
            else:
                response_data = None

        # check for server error
        if (req.status_code == 500):
            response_json = req.json()
            if ("error_message" in response_json):
                raise aurorax.AuroraXException("%s (%s)" % (response_json["error_message"],
                                                            req.status_code))
            else:
                raise aurorax.AuroraXException(response_json)

        # create reponse object
        res = AuroraXResponse(
            request=req, data=response_data, status_code=req.status_code)

        # return
        return res

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return pprint.pformat(self.__dict__)


class URLs:
    __DEFAULT_URL_DATA_SOURCES = "/api/v1/data_sources"
    __DEFAULT_URL_STATS = "/api/v1/stats"
    __DEFAULT_URL_EPHEMERIS_AVAILABILITY = "/api/v1/availability/ephemeris"
    __DEFAULT_URL_EPHEMERIS_UPLOAD = "/api/v1/data_sources/{}/ephemeris"
    __DEFAULT_URL_EPHEMERIS_SEARCH = "/api/v1/ephemeris/search"
    __DEFAULT_URL_EPHEMERIS_REQUEST = "/api/v1/ephemeris/requests/{}"
    __DEFAULT_URL_DATA_PRODUCTS_AVAILABILITY = "/api/v1/availability/data_products"
    __DEFAULT_URL_DATA_PRODUCTS_UPLOAD = "/api/v1/data_sources/{}/data_products"
    __DEFAULT_URL_DATA_PRODUCTS_SEARCH = "/api/v1/data_products/search"
    __DEFAULT_URL_DATA_PRODUCTS_REQUEST = "/api/v1/data_products/requests/{}"
    __DEFAULT_URL_CONJUNCTION_SEARCH = "/api/v1/conjunctions/search"
    __DEFAULT_URL_CONJUNCTION_REQUEST = "/api/v1/conjunctions/requests/{}"

    def __init__(self, base_url: str = DEFAULT_URL_BASE) -> None:
        self.__base = base_url
        self.__data_sources = self.__DEFAULT_URL_DATA_SOURCES
        self.__stats = self.__DEFAULT_URL_STATS
        self.__ephemeris_availability = self.__DEFAULT_URL_EPHEMERIS_AVAILABILITY
        self.__ephemeris_search = self.__DEFAULT_URL_EPHEMERIS_SEARCH
        self.__ephemeris_upload = self.__DEFAULT_URL_EPHEMERIS_UPLOAD
        self.__ephemeris_request = self.__DEFAULT_URL_EPHEMERIS_REQUEST
        self.__data_products_availability = self.__DEFAULT_URL_DATA_PRODUCTS_AVAILABILITY
        self.__data_products_search = self.__DEFAULT_URL_DATA_PRODUCTS_SEARCH
        self.__data_products_upload = self.__DEFAULT_URL_DATA_PRODUCTS_UPLOAD
        self.__data_products_request = self.__DEFAULT_URL_DATA_PRODUCTS_REQUEST
        self.__conjunction_search = self.__DEFAULT_URL_CONJUNCTION_SEARCH
        self.__conjunction_request = self.__DEFAULT_URL_CONJUNCTION_REQUEST

    @property
    def base_url(self) -> str:
        return self.__base

    @base_url.setter
    def base_url(self, value: str) -> None:
        self.__base = value

    # data sources
    # -------------------
    @property
    def data_sources_url(self) -> str:
        return "%s%s" % (self.__base, self.__data_sources)

    @property
    def stats_url(self) -> str:
        return "%s%s" % (self.__base, self.__stats)

    # availability
    # -------------------
    @property
    def ephemeris_availability_url(self) -> str:
        return "%s%s" % (self.__base, self.__ephemeris_availability)

    @property
    def data_products_availability_url(self) -> str:
        return "%s%s" % (self.__base, self.__data_products_availability)

    # ephemeris
    # -------------------
    @property
    def ephemeris_search_url(self) -> str:
        return "%s%s" % (self.__base, self.__ephemeris_search)

    @property
    def ephemeris_upload_url(self) -> str:
        return "%s%s" % (self.__base, self.__ephemeris_upload)

    @property
    def ephemeris_request_url(self) -> str:
        return "%s%s" % (self.__base, self.__ephemeris_request)

    # data products
    # -------------------
    @property
    def data_products_search_url(self) -> str:
        return "%s%s" % (self.__base, self.__data_products_search)

    @property
    def data_products_upload_url(self) -> str:
        return "%s%s" % (self.__base, self.__data_products_upload)

    @property
    def data_products_request_url(self) -> str:
        return "%s%s" % (self.__base, self.__data_products_request)

    # conjunctions
    # -------------------
    @property
    def conjunction_search_url(self) -> str:
        return f"{self.__base}{self.__conjunction_search}"

    @property
    def conjunction_request_url(self) -> str:
        return f"{self.__base}{self.__conjunction_request}"


# create instance of URLs that will be used throughout the application
urls = URLs()


def set_base_url(url: str) -> None:
    """
    Change the base URL for the API (ie. change to the staging system or local server).

    Attributes:
        url: new base url string (ie. 'https://api.staging.aurorax.space').

    """
    urls.base_url = url


def get_base_url() -> str:
    """
    Returns the current base URL for the API.
    """
    return urls.base_url


def reset_base_url() -> None:
    """
    Set the base URL for the API back to the default.
    """
    urls.base_url = DEFAULT_URL_BASE



"""
This module contains the Location class.
"""
from pydantic import BaseModel, validator
from typing import Union


class Location(BaseModel):
    """
    Class representing an AuroraX location (ie. geographic coordinates, GSM coordinates,
    northern/southern magnetic footprints).

    Attributes:
        lat: latitude (degrees East).
        lon: longitude (degrees North).

    """
    lat: Union[float, None]
    lon: Union[float, None]

    @validator("lon")
    def both_must_be_none_or_number(cls, v, values):
        if (v and not values["lat"]) or (values["lat"] and not v):
            raise ValueError(
                "lat and lon must both be numbers or both be None")
        return v

    def __str__(self) -> str:
        """
        String method.

        Returns:
            String format of Location object.

        """
        return str(self.__dict__)

    def __repr__(self) -> str:
        """
        Object representation.

        Returns:
            Object representation of Location object.

        """
        return "%s(lat=%s, lon=%s)" % (self.__class__.__name__, str(self.lat), str(self.lon))



"""
AuroraX holds ephemeris records for ground and space instruments in operation.
"""
import aurorax
import datetime
import humanize
import pprint
from pydantic import BaseModel
from typing import Dict, List, Optional


class Ephemeris(BaseModel):
    """
    Ephemeris data type.

    Attributes:
        data_source: aurorax.sources.DataSource source that the ephemeris record is associated with.
        epoch: datetime.datetime timestamp for the record in UTC.
        location_geo: aurorax.Location object with latitude and longitude in geographic coordinates.
        location_gsm: aurorax.Location object with latitude and longitude in GSM coordinates (leave empty for
            data sources with a type of 'ground').
        nbtrace: aurorax.Location object with north B-trace geomagnetic latitude and longitude.
        sbtrace: aurorax.Location object with south B-trace geomagnetic latitude and longitude.
        metadata: dictionary containing metadata values for this record.

    """
    data_source: aurorax.sources.DataSource
    epoch: datetime.datetime
    location_geo: aurorax.Location
    location_gsm: aurorax.Location = aurorax.Location(lat=None, lon=None)
    nbtrace: aurorax.Location
    sbtrace: aurorax.Location
    metadata: Dict = None

    def to_json_serializable(self) -> Dict:
        """
        Convert object to a JSON-serializable object (ie. translate datetime objects to strings).

        Returns:
            Dictionary JSON-serializable object.

        """
        d = self.__dict__

        # format epoch as str
        if (type(d["epoch"]) is datetime.datetime):
            d["epoch"] = d["epoch"].strftime("%Y-%m-%dT%H:%M:00.000Z")

        # format location
        if (type(d["location_geo"]) is aurorax.Location):
            d["location_geo"] = d["location_geo"].__dict__
        if (type(d["location_gsm"]) is aurorax.Location):
            d["location_gsm"] = d["location_gsm"].__dict__
        if (type(d["nbtrace"]) is aurorax.Location):
            d["nbtrace"] = d["nbtrace"].__dict__
        if (type(d["sbtrace"]) is aurorax.Location):
            d["sbtrace"] = d["sbtrace"].__dict__

        # format metadata
        if (type(self.metadata) is dict):
            for key, value in self.metadata.items():
                if (type(value) is datetime.datetime or type(value) is datetime.date):
                    self.metadata[key] = self.metadata[key].strftime(
                        "%Y-%m-%dT%H:%M:%S.%f")
        if (type(self.metadata) is list):
            self.metadata = {}

        # format data source fields for query
        d["program"] = self.data_source.program
        d["platform"] = self.data_source.platform
        d["instrument_type"] = self.data_source.instrument_type
        del d["data_source"]

        # return
        return d

    def __str__(self) -> str:
        """
        String method.

        Returns:
            String format of Ephemeris.

        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation.

        Returns:
            Object representation of Ephemeris.

        """
        return pprint.pformat(self.__dict__)


class Search():
    """
    Class representing an AuroraX ephemeris search.

    At least one search criteria from programs, platforms, instrument_types, or metadata_filters
    must be specified.

    start: start datetime.datetime timestamp of the search.
    end: end datetime.datetime timestamp of the search.
    programs: list of program names to search.
    platforms: list of platform names to search.
    instrument_types: list of instrument types to search.
    metadata_filters: list of dictionaries describing metadata keys and values to filter on, defaults to None.
        e.g. {
            "key": "string",
            "operator": "=",
            "values": [
                "string"
            ]
        }
    request: aurorax.AuroraXResponse object returned when the search is executed.
    request_id: unique AuroraX string ID assigned to the request.
    request_url: unique AuroraX URL string assigned to the request.
    executed: boolean, gets set to True when the search is executed.
    completed: boolean, gets set to True when the search is checked to be finished.
    data_url: URL string where data is accessed.
    query: dictionary of values sent for the search query.
    status: dictionary of status updates.
    data: list of aurorax.ephemeris.Ephemeris objects returned.
    logs: list of logging messages from the API.
    """

    def __init__(self,
                 start: datetime.datetime,
                 end: datetime.datetime,
                 programs: List[str] = None,
                 platforms: List[str] = None,
                 instrument_types: List[str] = None,
                 metadata_filters: List[Dict] = None) -> None:
        """
        Create a new Search object.

        """
        self.request = None
        self.request_id = ""
        self.request_url = ""
        self.executed = False
        self.completed = False
        self.data_url = ""
        self.query = {}
        self.status = {}
        self.data: List[Ephemeris] = []
        self.logs = []

        self.start = start
        self.end = end
        self.programs = programs
        self.platforms = platforms
        self.instrument_types = instrument_types
        self.metadata_filters = metadata_filters

    def __str__(self) -> str:
        """
        String method.

        Returns:
            String format of Ephemeris Search object.

        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation.

        Returns:
            Object representation of Ephemeris Search object.

        """
        return pprint.pformat(self.__dict__)

    def execute(self) -> None:
        """
        Initiate ephemeris search request.

        Raises:
            aurorax.exceptions.AuroraXBadParametersException: missing parameters.

        """
        # check for at least one filter criteria
        if not (self.programs or self.platforms or self.instrument_types or self.metadata_filters):
            raise aurorax.AuroraXBadParametersException(
                "At least one filter criteria parameter besides Start and End must be specified.")

        # set up request
        url = aurorax.api.urls.ephemeris_search_url
        post_data = {
            "data_sources": {
                "programs": [] if not self.programs else self.programs,
                "platforms": [] if not self.platforms else self.platforms,
                "instrument_types": [] if not self.instrument_types else self.instrument_types,
                "ephemeris_metadata_filters": [] if not self.metadata_filters else self.metadata_filters,
            },
            "start": self.start.strftime("%Y-%m-%dT%H:%M:%S"),
            "end": self.end.strftime("%Y-%m-%dT%H:%M:%S"),
        }
        self.query = post_data

        # do request
        req = aurorax.AuroraXRequest(
            method="post", url=url, body=post_data, null_response=True)
        res = req.execute()

        # set request ID, request_url, executed
        self.executed = True
        if (res.status_code == 202):
            # request successfully dispatched
            self.executed = True
            self.request_url = res.request.headers["location"]
            self.request_id = self.request_url.rsplit("/", 1)[-1]
        self.request = res

    def update_status(self, status: Dict = None) -> None:
        """
        Update the status of this ephemeris search request.

        Attributes:
            status: retrieved status dictionary (include to avoid requesting it from the API again), defaults to None.
        """
        # get the status if it isn't passed in
        if (status is None):
            status = aurorax.requests.get_status(self.request_url)

        # update request status by checking if data URI is set
        if (status["search_result"]["data_uri"] is not None):
            self.completed = True
            self.data_url = "%s%s" % (
                aurorax.api.urls.base_url, status["search_result"]["data_uri"])

        # set class variable "status" and "logs"
        self.status = status
        self.logs = status["logs"]

    def check_for_data(self) -> bool:
        """
        Check to see if data is available for this ephemeris search request.

        Returns:
            True if data is available, else False.
        """
        self.update_status()

        return self.completed

    def get_data(self) -> None:
        """
        Retrieve the data available for this ephemeris search request.
        """
        if (self.completed is False):
            print("No data available, update status or check for data first")
            return
        url = self.data_url
        raw_data = aurorax.requests.get_data(url)

        self.data = [Ephemeris(**e) for e in raw_data]

    def wait(self, poll_interval: float = aurorax.requests.STANDARD_POLLING_SLEEP_TIME, verbose: bool = False) -> None:
        """
        Block and wait for the request to complete and data is available for retrieval.

        Attributes:
            poll_interval: time in seconds to wait between polling attempts,
                defaults to aurorax.requests.STANDARD_POLLING_SLEEP_TIME.
            verbose: output poll times, defaults to False.

        """
        url = aurorax.api.urls.ephemeris_request_url.format(self.request_id)
        self.update_status(aurorax.requests.wait_for_data(
            url, poll_interval=poll_interval, verbose=verbose))

    def cancel(self, wait: bool = False, verbose: bool = False,
               poll_interval: float = aurorax.requests.STANDARD_POLLING_SLEEP_TIME) -> int:
        """
        Cancel the ephemeris search request at the API. This method returns asynchronously by default.

        Attributes:
            wait: set to True to block until the cancellation request has been completed. This may take several minutes.
            verbose: when wait=True, output poll times, defaults to False.
            poll_interval: when wait=True, seconds to wait between polling calls, defaults to STANDARD_POLLING_SLEEP_TIME.

        Returns:
            1 on success.

        Raises:
            aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.
            aurorax.exceptions.AuroraXUnauthorizedException: invalid API key for this operation.

        """
        url = aurorax.api.urls.ephemeris_request_url.format(self.request_id)
        return aurorax.requests.cancel(url, wait=wait, poll_interval=poll_interval, verbose=verbose)


def search_async(start: datetime.datetime,
                 end: datetime.datetime,
                 programs: List[str] = None,
                 platforms: List[str] = None,
                 instrument_types: List[str] = None,
                 metadata_filters: List[Dict] = None) -> Search:
    """
    Submit a request for an ephemeris search, return asynchronously.

    At least one search criteria from programs, platforms, instrument_types, or metadata_filters
    must be specified.

    Attributes:
        start: start datetime.datetime timestamp of the search.
        end: end datetime.datetime timestamp of the search.
        programs: list of programs to search through, defaults to None.
        platforms: list of platforms to search through, defaults to None.
        instrument_types: list of instrument types to search through, defaults to None.
        metadata_filters: list of dictionaries describing metadata keys and values to filter on, defaults to None.
            e.g. {
                "key": "string",
                "operator": "=",
                "values": [
                    "string"
                ]
            }

    Returns:
        An aurorax.ephemeris.Search object.

    Raises:
        aurorax.exceptions.AuroraXBadParametersException: missing parameters.

    """
    s = aurorax.ephemeris.Search(start=start,
                                 end=end,
                                 programs=programs,
                                 platforms=platforms,
                                 instrument_types=instrument_types,
                                 metadata_filters=metadata_filters)
    s.execute()
    return s


def search(start: datetime.datetime,
           end: datetime.datetime,
           programs: List[str] = None,
           platforms: List[str] = None,
           instrument_types: List[str] = None,
           metadata_filters: List[Dict] = None,
           verbose: bool = False,
           poll_interval: float = aurorax.requests.STANDARD_POLLING_SLEEP_TIME) -> Search:
    """
    Search for ephemeris records.

    At least one search criteria from programs, platforms, instrument_types, or metadata_filters
    must be specified.

    Attributes:
        start: start datetime.datetime timestamp of the search.
        end: end datetime.datetime timestamp of the search.
        programs: list of programs to search through, defaults to None.
        platforms: list of platforms to search through, defaults to None.
        instrument_types: list of instrument types to search through, defaults to None.
        metadata_filters: list of dictionaries describing metadata keys and values to filter on, defaults to None.
            e.g. {
                "key": "string",
                "operator": "=",
                "values": [
                    "string"
                ]
            }
        verbose: output poll times, defaults to False.
        poll_interval: time in seconds to wait between polling attempts,
            defaults to aurorax.requests.STANDARD_POLLING_SLEEP_TIME.

    Returns:
        An aurorax.ephemeris.Search object.

    Raises:
        aurorax.exceptions.AuroraXBadParametersException: missing parameters

    """
    # create a Search() object
    s = Search(start=start, end=end, programs=programs, platforms=platforms,
               instrument_types=instrument_types, metadata_filters=metadata_filters)
    if (verbose is True):
        print("[%s] Search object created" % (datetime.datetime.now()))

    # execute the search
    s.execute()
    if (verbose is True):
        print("[%s] Request submitted" % (datetime.datetime.now()))
        print("[%s] Request ID: %s" % (datetime.datetime.now(), s.request_id))
        print("[%s] Request details available at: %s" %
              (datetime.datetime.now(), s.request_url))

    # wait for data
    if (verbose is True):
        print("[%s] Waiting for data ..." % (datetime.datetime.now()))
    s.wait(poll_interval=poll_interval, verbose=verbose)

    # get the data
    if (verbose is True):
        print("[%s] Retrieving data ..." % (datetime.datetime.now()))
    s.get_data()

    # return response with the data
    if (verbose is True):
        print("[%s] Retrieved %s of data containing %d records" % (datetime.datetime.now(),
                                                                   humanize.filesize.naturalsize(
                                                                       s.status["search_result"]["file_size"]),
                                                                   s.status["search_result"]["result_count"]))
    return s


def __validate_data_source(identifier: int, records: List[Ephemeris]) -> Optional[Ephemeris]:
    # get all current sources
    sources = {source.identifier: source for source in aurorax.sources.list()}
    if identifier not in sources.keys():
        raise aurorax.AuroraXValidationException(
            f"Data source with unique identifier {identifier} could not be found.")

    for record in records:
        # check the identifier, program name, platform name, and instrument type
        try:
            reference = sources[record.data_source.identifier]
        except KeyError:
            raise aurorax.AuroraXValidationException(
                f"Data source with unique identifier {record.data_source.identifier} could not be found.")

        if not (record.data_source.program == reference.program
                and record.data_source.platform == reference.platform
                and record.data_source.instrument_type == reference.instrument_type):
            return record

    return None


def upload(identifier: int, records: List[Ephemeris], validate_source: bool = False) -> int:
    """
    Upload ephemeris records to AuroraX.

    Attributes:
        identifier: AuroraX data source ID int.
        records: list of aurorax.ephemeris.Ephemeris records to upload.
        validate_source: boolean, set to True to validate all records before uploading.

    Returns:
        1 for success, raises exception on error.

    Raises:
        aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected content error.
        aurorax.exceptions.AuroraXUploadException: upload error.
        aurorax.exceptions.AuroraXValidationException: data source validation error.

    """
    # validate record sources if the flag is set
    if validate_source:
        validation_error = __validate_data_source(identifier, records)
        if validation_error:
            raise aurorax.AuroraXValidationException(
                "Unable to validate data source found in record: {}".format(validation_error))

    # translate each ephemeris record to a request-friendly dict (ie. convert datetimes to strings, etc.)
    for i, _ in enumerate(records):
        if (type(records[i]) is Ephemeris):
            records[i] = records[i].to_json_serializable()

    # make request
    url = aurorax.api.urls.ephemeris_upload_url.format(identifier)
    req = aurorax.AuroraXRequest(
        method="post", url=url, body=records, null_response=True)
    res = req.execute()

    # evaluate response
    if (res.status_code == 400):
        raise aurorax.AuroraXUploadException(
            "%s - %s" % (res.data["error_code"], res.data["error_message"]))

    # return
    return 1


def delete(data_source: aurorax.sources.DataSource, start: datetime.datetime, end: datetime.datetime) -> int:
    """
    Delete a range of ephemeris records. This method is asynchronous.

    Attributes:
        data_source: aurorax.sources.DataSource source associated with the data product records.
            Identifier, program, platform, and instrument_type are required.
        start: datetime.datetime beginning of range to delete records for, inclusive.
        end: datetime.datetime end of datetime range to delete records for, inclusive.

    Returns:
        1 on success.

    Raises:
        aurorax.exceptions.AuroraXMaxRetriesException: max retry error.
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.
        aurorax.exceptions.AuroraXNotFoundException: source not found.
        aurorax.exceptions.AuroraXUnauthorizedException: invalid API key for this operation.
        aurorax.exceptions.AuroraXBadParametersException: missing parameters.

    """
    if not all([data_source.identifier, data_source.program, data_source.platform, data_source.instrument_type]):
        raise aurorax.AuroraXBadParametersException(
            "One or more required data source parameters are missing. Delete operation aborted.")

    # do request
    url = aurorax.api.urls.ephemeris_upload_url.format(data_source.identifier)
    params = {
        "program": data_source.program,
        "platform": data_source.platform,
        "instrument_type": data_source.instrument_type,
        "start": start.strftime("%Y-%m-%dT%H:%M:%S"),
        "end": end.strftime("%Y-%m-%dT%H:%M:%S")
    }
    delete_req = aurorax.AuroraXRequest(
        method="delete", url=url, body=params, null_response=True)
    res = delete_req.execute()

    # evaluate response
    if (res.status_code == 400):
        if type(res.data) is list:
            raise aurorax.AuroraXBadParametersException(
                "%s - %s" % (res.status_code, res.data[0]["message"]))
        raise aurorax.AuroraXBadParametersException(
            "%s - %s" % (res.data["error_code"], res.data["error_message"]))

    # return
    return 1



"""
The exceptions module contains descriptive exceptions unique to AuroraX.
"""


class AuroraXException(Exception):

    def __init__(self, *args, **kwargs):
        response = kwargs.pop("response", None)
        self.response = response
        self.request = kwargs.pop("request", None)
        if (response is not None and not self.request and hasattr(response, "request")):
            self.request = self.response.request
        super(AuroraXException, self).__init__(*args, **kwargs)


class AuroraXNotFoundException(AuroraXException):
    """
    The AuroraX record was not found.
    """
    pass


class AuroraXDuplicateException(AuroraXException):
    """
    A duplicate record already exists.
    """
    pass


class AuroraXValidationException(AuroraXException):
    """
    Validation of data failed.
    """
    pass


class AuroraXUnexpectedContentTypeException(AuroraXException):
    """
    The API responded with an unexpected content type.
    """
    pass


class AuroraXMaxRetriesException(AuroraXException):
    """
    The maximum number of retries for the request has been reached.
    """
    pass


class AuroraXBadParametersException(AuroraXException):
    """
    Bad parameters were given in the request.
    """
    pass


class AuroraXUnauthorizedException(AuroraXException):
    """
    A privileged operation was attempted without authorization.
    """
    pass


class AuroraXConflictException(AuroraXException):
    """
    A conflict occurred while modifying records.
    """
    pass


class AuroraXUploadException(AuroraXException):
    """
    Error occurred during upload operation.
    """
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aurorax', 'aurorax._internal']

package_data = \
{'': ['*']}

install_requires = \
['aacgmv2>=2.6.2,<3.0.0',
 'flake8>=3.8.4,<4.0.0',
 'humanize>=3.2.0,<4.0.0',
 'pydantic>=1.7.3,<2.0.0',
 'requests>=2.25.1,<3.0.0']

setup_kwargs = {
    'name': 'pyaurorax',
    'version': '0.6.11',
    'description': 'Python library for interacting with the AuroraX API',
    'long_description': '# PyAuroraX\n\n[![Github Actions - Tests](https://github.com/aurorax-space/pyaurorax/workflows/tests/badge.svg)](https://github.com/aurorax-space/pyaurorax/actions?query=workflow%3Atests)\n[![PyPI version](https://img.shields.io/pypi/v/pyaurorax.svg)](https://pypi.python.org/pypi/pyaurorax/)\n[![MIT license](https://img.shields.io/badge/License-MIT-blue.svg)](https://lbesson.mit-license.org/)\n[![PyPI Python versions](https://img.shields.io/pypi/pyversions/pyaurorax.svg)](https://pypi.python.org/pypi/pyaurorax/)\n\nPython library for interacting with the AuroraX API.\n\n## Installing PyAuroraX\n\nPyAuroraX is available on PyPI:\n\n```console\n$ python -m pip install pyaurorax\n```\n\n## Supported Python Versions\n\nPyAuroraX officially supports Python 3.6+.\n\n## Usage\n\n```python\n>>> import aurorax\n```\n\n## Development\n\nClone the repository and install dependencies using Poetry.\n\n```console\n$ git clone https://github.com/aurorax-space/pyaurorax.git\n$ cd pyaurorax\n$ make install\n```\n\n## Testing\n\n```console\n$ make test\n[ or do each test separately ]\n$ make test-flake8\n$ make test-pylint\n$ make test-bandit\n$ make test-pytest\n```\n\n## Additional Testing for Development Environments\n\nTo run additional tests that are not integrated into the CI pipeline, run the following:\n\n```console\n$ make test-additional\n```\n',
    'author': 'Darren Chaddock',
    'author_email': 'dchaddoc@ucalgary.ca',
    'maintainer': 'Darren Chaddock',
    'maintainer_email': 'dchaddoc@ucalgary.ca',
    'url': 'https://github.com/aurorax-space/pyaurorax',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
The PyAuroraX package provides a way to interact with the [AuroraX REST API](https://aurorax.space/data/api_libraries).
It is intended to provide an intuitive process for those in the space physics and citizen science communities to
programmatically query AuroraX's vast database for conjunctions, ephemeris or data product records, data availability, and
statistics. Requires Python3.6+.

Check out this project on [GitHub](https://github.com/aurorax-space/pyaurorax) and explore the evolving ecosystem of
visualizations, tools, and data at [AuroraX](https://aurorax.space/).

For an overview of intended usage and examples, visit the
[AuroraX Documentation website](https://docs.aurorax.space/python_libraries/pyaurorax/).
Details of functionality and options are available in the
[API reference](https://docs.aurorax.space/python_libraries/pyaurorax/api_reference/aurorax/).

Installation:
```console
$ pip install pyaurorax
```

Basic usage:
```python
> import aurorax
```
"""
__version__ = "0.6.11"

# pull in top level functions
from .api import AuroraXRequest
from .api import authenticate
from .api import get_api_key

# pull in exceptions at top level
from .exceptions import AuroraXException
from .exceptions import AuroraXNotFoundException
from .exceptions import AuroraXMaxRetriesException
from .exceptions import AuroraXDuplicateException
from .exceptions import AuroraXUnexpectedContentTypeException
from .exceptions import AuroraXValidationException
from .exceptions import AuroraXBadParametersException
from .exceptions import AuroraXUnauthorizedException
from .exceptions import AuroraXConflictException
from .exceptions import AuroraXUploadException

# pull in models
from .models import Location

# pull in modules
from aurorax import requests
from aurorax import exceptions
from aurorax import api
from aurorax import sources
from aurorax import metadata
from aurorax import availability
from aurorax import ephemeris
from aurorax import data_products
from aurorax import conjunctions
from aurorax import util



"""
AuroraX metadata schemas describe the intended structure of metadata stored in
ephemeris and data product records.
"""
import aurorax
from typing import Dict, List


def validate(schema: List[Dict], record: Dict, quiet: bool = False) -> bool:
    """
    Validate the metadata record against a schema. This checks that the key names match and there aren't
    fewer or more keys than expected.

    Attributes:
        schema: list of dictionaries representing the metadata schema to validate against.
        record: metadata record dictionary to validate.

    Returns:
        True if the metadata record is valid.
    """
    # check keys
    schema_keys = sorted([i["field_name"] for i in schema])
    record_keys = sorted(record.keys())
    if (schema_keys != record_keys):
        if (quiet is False):
            print("Metadata invalid, keys don't match")
        return False
    return True


def get_ephemeris_schema(identifier: int) -> List[Dict]:
    """
    Retrieve the metadata schema for an ephemeris record.

    Attributes:
        identifier: ephemeris source ID.

    Returns:
        Metadata schema associated with the record.

    """
    source_info = aurorax.sources.get_using_identifier(
        identifier, format="full_record")
    if source_info.ephemeris_metadata_schema:
        return source_info.ephemeris_metadata_schema
    else:
        return []


def get_data_products_schema(identifier: int) -> List[Dict]:
    """
    Retrieve the metadata schema for a data products record.

    Attributes:
        identifier: ephemeris source ID.

    Returns:
        Metadata schema associated with the record.

    """
    source_info = aurorax.sources.get_using_identifier(
        identifier, format="full_record")
    if source_info.data_product_metadata_schema:
        return source_info.data_product_metadata_schema
    else:
        return []



"""
The requests module contains methods for retrieving data from an AuroraX request.
"""
import aurorax
import datetime
import time
from typing import Dict, List

# globals
FIRST_FOLLOWUP_SLEEP_TIME = 0.050  # 50ms
STANDARD_POLLING_SLEEP_TIME = 1.0  # 1s


def get_status(request_url: str) -> Dict:
    """
    Retrieve the status of a request.

    Attributes:
        request_url: URL of the request information.

    Returns:
        Status dictionary for the request.

    """
    # do request
    req = aurorax.AuroraXRequest(method="get", url=request_url)
    res = req.execute()

    # return
    return res.data


def get_data(data_url: str) -> List:
    """
    Retrieve the data for a request.

    Attributes:
        data_url: URL for the data of a request.

    Returns:
        List of JSON data objects in the response.

    """
    # do request
    req = aurorax.AuroraXRequest(method="get", url=data_url)
    res = req.execute()

    # set data var
    data_result = res.data["result"]

    # serialize epochs and locations into datetimes and Locations
    for i in range(0, len(data_result)):
        if ("epoch" in data_result[i]):
            data_result[i]["epoch"] = datetime.datetime.strptime(data_result[i]["epoch"],
                                                                 "%Y-%m-%dT%H:%M:%S")
        if ("location_geo" in data_result[i]):
            data_result[i]["location_geo"] = aurorax.Location(lat=data_result[i]["location_geo"]["lat"],
                                                              lon=data_result[i]["location_geo"]["lon"])
        if ("location_gsm" in data_result[i]):
            data_result[i]["location_gsm"] = aurorax.Location(lat=data_result[i]["location_gsm"]["lat"],
                                                              lon=data_result[i]["location_gsm"]["lon"])
        if ("nbtrace" in data_result[i]):
            data_result[i]["nbtrace"] = aurorax.Location(lat=data_result[i]["nbtrace"]["lat"],
                                                         lon=data_result[i]["nbtrace"]["lon"])
        if ("sbtrace" in data_result[i]):
            data_result[i]["sbtrace"] = aurorax.Location(lat=data_result[i]["sbtrace"]["lat"],
                                                         lon=data_result[i]["sbtrace"]["lon"])

    # return
    return data_result


def get_logs(request_url: str) -> List:
    """
    Retrieve the logs for a request.

    Attributes:
        request_url: URL of the request information.

    Returns:
        List of logged messages for the request.

    """
    # get status
    status = get_status(request_url)

    if ("logs" in status):
        return status["logs"]
    else:
        return []


def wait_for_data(request_url: str,
                  poll_interval: float = STANDARD_POLLING_SLEEP_TIME,
                  verbose: bool = False) -> Dict:
    """
    Block and wait for the data to be made available for a request.

    Attributes:
        request_url: URL of the request information.
        poll_interval: seconds to wait between polling calls, defaults to STANDARD_POLLING_SLEEP_TIME.
        verbose: output poll times, defaults to False.

    Returns:
        Status dictionary for the request.

    """
    status = get_status(request_url)
    while (status["search_result"]["data_uri"] is None):
        time.sleep(poll_interval)
        if (verbose is True):
            print("[%s] Checking for data ..." % (datetime.datetime.now()))
        status = get_status(request_url)
    if (verbose is True):
        print("[%s] Data is now available" % (datetime.datetime.now()))
    return status


def cancel(request_url: str,
           wait: bool = False,
           poll_interval: float = STANDARD_POLLING_SLEEP_TIME,
           verbose: bool = False) -> int:
    """
    Cancel the request at the given URL. This operation is asynchronous by default unless the wait param is set to True.

    Attributes:
        request_url: URL string of the request to be canceled.
        wait: set to True to block until the cancellation request has been completed. This may take several minutes.
        verbose: when wait=True, output poll times, defaults to False.
        poll_interval: when wait=True, seconds to wait between polling calls, defaults to STANDARD_POLLING_SLEEP_TIME.

    Returns:
        1 on success.

    Raises:
        aurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error.
        aurorax.exceptions.AuroraXUnauthorizedException: invalid API key for this operation.

    """

    # do request
    req = aurorax.AuroraXRequest(
        method="delete", url=request_url, null_response=True)
    req.execute()

    if not wait:
        return 1

    status = get_status(request_url)
    while (status["search_result"]["data_uri"] is None and status["search_result"]["error_condition"] is False):
        time.sleep(poll_interval)
        if (verbose is True):
            print("[%s] Checking for cancellation status ..." %
                  (datetime.datetime.now()))
        status = get_status(request_url)
    if (verbose is True):
        print("[%s] The request has been cancelled" %
              (datetime.datetime.now()))

    return 1
